#include "esp_camera.h"
#include "camera_def.h"
#include <base64.h> 
#include <AcksenButton.h>

#include <WiFi.h>
#include <WiFiManager.h>  // https://github.com/tzapu/WiFiManager 
#include <esp_wifi.h>

#include <mqtt.h>

// use 12 bit precision for LEDC timer
#define LEDC_TIMER_12_BIT 12

// use 5000 Hz as a LEDC base frequency
#define LEDC_BASE_FREQ 5000

#define BUTTON_OK_GPIO_NUM 47

#define CONFIG_LED_ILLUMINATOR_ENABLED 1
#define FLASH_GPIO_NUM 41
int led_duty = 180;

AcksenButton OK_Button(BUTTON_OK_GPIO_NUM, ACKSEN_BUTTON_MODE_NORMAL, 42, INPUT_PULLUP);

void ARDUINO_ISR_ATTR action() {
  Serial.println("interrupt");
}

void setup() {
  Serial.begin(115200);

  pinMode(10, INPUT_PULLUP);
  // digitalWrite(10, LOW); 

  // startIncuWifi();
  // reconnMqtt();
  // initCamera();



    // digitalWrite(HEATER_GPIO_NUM, LOW);   

  # if defined(CAMERA_MODEL_ESP32S3_EYE)
    Serial.println("CAMERA_MODEL_ESP32S3_EYE");
  #endif

  if(ledcAttach(FLASH_GPIO_NUM, LEDC_BASE_FREQ, LEDC_TIMER_12_BIT)) {
    Serial.println("LEDC attached");
  } else {
    Serial.println("LEDC failed");
  }

  // pinMode(HEATER_GPIO_NUM, OUTPUT);
  // digitalWrite(HEATER_GPIO_NUM, LOW);   

  // ledcWrite(FLASH_GPIO_NUM, 21);
}

WiFiManager wm;
bool startIncuWifi() {
  bool res = wm.autoConnect("AutoConnectAP");  // password protected ap

  if (!res) {
    Serial.println("Failed to connect wifi");
  } else {
    Serial.println("wifi connected...yeey :)");
  }

  return res;
}

void loop() {
  // put your main code here, to run repeatedly:

  if (OK_Button.onPressed()) {
    // digitalWrite(HEATER_GPIO_NUM, measure);  
    Serial.println("OK");

    // detachInterrupt(10);

    initCamera();
    sendImage(); 

    // esp_camera_return_all();
    esp_camera_deinit();
    // deinit_cam();

    // pinMode(6, OUTPUT);
  // pinMode(10, INPUT_PULLUP);
  // attachInterrupt(10, action, FALLING);

    Serial.print("PWDN_GPIO_NUM pin is: ");
    Serial.println(PWDN_GPIO_NUM);

    digitalWrite(PWDN_GPIO_NUM, LOW);   
    digitalWrite(PWDN_GPIO_NUM, HIGH);   

  }

  // mqttLoop();

  OK_Button.refreshStatus();
}

void deinit_cam() {
    esp_camera_deinit();

    camera_config_t config;
	config.ledc_channel = LEDC_CHANNEL_0;
	config.ledc_timer = LEDC_TIMER_0;
	config.pin_d0 = -1;
	config.pin_d1 = -1;
	config.pin_d2 = -1;
	config.pin_d3 = -1;
	config.pin_d4 = -1;
	config.pin_d5 = -1;
	config.pin_d6 = -1;
	config.pin_d7 = -1;
	config.pin_xclk = -1;
	config.pin_pclk = -1;
	config.pin_vsync = -1;
	config.pin_href = -1;
	config.pin_sscb_sda = -1;
	config.pin_sscb_scl = -1;
	config.pin_pwdn = -1;
	config.pin_reset = -1;
	config.xclk_freq_hz = 20000000;
	config.pixel_format = PIXFORMAT_JPEG;

  esp_err_t err = esp_camera_init(&config);
	if (err != ESP_OK)
	{
		Serial.printf("Camera init failed with error 0x%x", err);
	} else {
    Serial.println("deinit scss!");
  }
}

bool initCamera() {
  camera_config_t config;
	config.ledc_channel = LEDC_CHANNEL_0;
	config.ledc_timer = LEDC_TIMER_0;
	config.pin_d0 = Y2_GPIO_NUM;
	config.pin_d1 = Y3_GPIO_NUM;
	config.pin_d2 = Y4_GPIO_NUM;
	config.pin_d3 = Y5_GPIO_NUM;
	config.pin_d4 = Y6_GPIO_NUM;
	config.pin_d5 = Y7_GPIO_NUM;
	config.pin_d6 = Y8_GPIO_NUM;
	config.pin_d7 = Y9_GPIO_NUM;
	config.pin_xclk = XCLK_GPIO_NUM;
	config.pin_pclk = PCLK_GPIO_NUM;
	config.pin_vsync = VSYNC_GPIO_NUM;
	config.pin_href = HREF_GPIO_NUM;
	config.pin_sscb_sda = SIOD_GPIO_NUM;
	config.pin_sscb_scl = SIOC_GPIO_NUM;
	config.pin_pwdn = PWDN_GPIO_NUM;
	config.pin_reset = RESET_GPIO_NUM;
	config.xclk_freq_hz = 20000000;
	config.pixel_format = PIXFORMAT_JPEG;
	//init with high specs to pre-allocate larger buffers
	if (psramFound())
	{
    Serial.println("\nFound PSRAM, this will improve performance!");
		config.frame_size = FRAMESIZE_XGA;
		config.jpeg_quality = 9;
		config.fb_count = 2;
    config.grab_mode = CAMERA_GRAB_LATEST;
	}
	else
	{
    Serial.println("\nWarning, PSRAM not detected. If your board has PSRAM, activate it from Tools/PSRAM ('OPI PRAM' option worked for me) !");
		config.frame_size = FRAMESIZE_SVGA;
		config.jpeg_quality = 12;
		config.fb_count = 1;
	}

#if defined(CAMERA_MODEL_ESP_EYE)
	pinMode(13, INPUT_PULLUP);
	pinMode(14, INPUT_PULLUP);
#endif

	// camera init
	esp_err_t err = esp_camera_init(&config);
	if (err != ESP_OK)
	{
		Serial.printf("Camera init failed with error 0x%x", err);
		return false;
	}

	sensor_t *s = esp_camera_sensor_get();
	//initial sensors are flipped vertically and colors are a bit saturated
	if (s->id.PID == OV3660_PID)
	{
		s->set_vflip(s, 1);		  //flip it back
		s->set_brightness(s, 1);  //up the blightness just a bit
		s->set_saturation(s, -2); //lower the saturation
	}
	//drop down frame size for higher initial frame rate
	// s->set_framesize(s, FRAMESIZE_UXGA);

  

#if defined(CAMERA_MODEL_M5STACK_WIDE)
	s->set_vflip(s, 1);
	s->set_hmirror(s, 1);
#endif
  return true;
}

void sendImage() {
  camera_fb_t * fb = NULL;

  fb = esp_camera_fb_get();
  esp_camera_fb_return(fb);

#if CONFIG_LED_ILLUMINATOR_ENABLED
  Serial.println("CONFIG_LED_ILLUMINATOR_ENABLED");
  enable_led(true);
  vTaskDelay(150 / portTICK_PERIOD_MS);  // The LED needs to be turned on ~150ms before the call to esp_camera_fb_get()
  fb = esp_camera_fb_get();              // or it won't be visible in the frame. A better way to do this is needed.
  enable_led(false);
#else
  fb = esp_camera_fb_get();
#endif
  
  if(!fb) {
    Serial.println("Camera capture failed");
    return;
  }  

  Serial.println("Camera capture success");

  // uint8_t *input = fb->buf;
  String imageData = "data:image/jpeg;base64,";
  // for (size_t i = 0; i < fb->len; i++) {
  //   imageData += (char)fb->buf[i];
  // }

    // auto inputLength = sizeof(fb->buf);
    // char output[base64::encodeLength(fb->len)];
    // base64::encode(fb->buf, sizeof(fb->buf), output);

    imageData += base64::encode(fb->buf, fb->len);;

  // imageData += String(output);


  // char output[base64::encodeLength(3)];
  // String imageFile = "data:image/jpeg;base64," + baseImgString;
  // int inputStringLength = strlen(imageFile);

  // for (int i=0; i<fb->len; i++) {
  //   base64::encode((input++), 3, output);
  //   if (i%3==0) imageFile += String(output);
  // }

  // int encodedLength = Base64.encodedLength(inputStringLength);
  // char encodedString[encodedLength + 1];
  // Base64.encode(encodedString, input, inputStringLength);

  // int fbLen = imageFile.length();

  Serial.print("publishing img of length: ");
  Serial.println(imageData.length());
  
  // publishImage(imageData.length(), imageData);
  
  esp_camera_fb_return(fb);
  // return "failed, rc="+client.state();
}

void enable_led(bool en) {  // Turn LED On or Off
  Serial.println("begin ledc");

  int duty = en ? led_duty : 0;
  log_i("Set LED intensity to %d", duty);

  ledcWrite(FLASH_GPIO_NUM, duty);
  // ledc_set_duty(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_0, duty);
  // ledc_update_duty(LEDC_LOW_SPEED_MODE, LEDC_CHANNEL_0);

  log_i("Set LED intensity to %d", duty);
  Serial.println("end ledc");
}
